###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import asyncio
import copy
import math
import struct
from typing import Optional

import txaio

from autobahn.asyncio.util import (
    create_transport_details,
    get_serializers,
    transport_channel_id,
)
from autobahn.util import _LazyHexFormatter, hltype, public
from autobahn.wamp.exception import ProtocolError, SerializationError, TransportLost
from autobahn.wamp.types import TransportDetails

__all__ = (
    "WampRawSocketClientFactory",
    "WampRawSocketClientProtocol",
    "WampRawSocketServerFactory",
    "WampRawSocketServerProtocol",
)

FRAME_TYPE_DATA = 0
FRAME_TYPE_PING = 1
FRAME_TYPE_PONG = 2

MAGIC_BYTE = 0x7F


class PrefixProtocol(asyncio.Protocol):
    prefix_format = "!L"
    prefix_length = struct.calcsize(prefix_format)
    max_length = 16 * 1024 * 1024
    max_length_send = max_length
    log = txaio.make_logger()  # @UndefinedVariable
    peer: str | None = None
    is_server: bool | None = None

    @property
    def transport_details(self) -> TransportDetails | None:
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.transport_details`
        """
        return self._transport_details

    def connection_made(self, transport):
        # asyncio networking framework entry point, called by asyncio
        # when the connection is established (either a client or a server)
        self.log.debug(
            "RawSocker Asyncio: Connection made with peer {peer}", peer=self.peer
        )

        self.transport = transport

        # determine preliminary transport details (what is known at this point)
        self._transport_details = create_transport_details(
            self.transport, self.is_server
        )
        self._transport_details.channel_framing = (
            TransportDetails.CHANNEL_FRAMING_RAWSOCKET
        )

        # backward compatibility
        self.peer = self._transport_details.peer

        self._buffer = b""
        self._header = None
        self._wait_closed = txaio.create_future()

    @property
    def is_closed(self):
        if hasattr(self, "_wait_closed"):
            return self._wait_closed
        else:
            f = txaio.create_future()
            f.set_result(True)
            return f

    def connection_lost(self, exc):
        self.log.debug("RawSocker Asyncio: Connection lost")
        self.transport = None
        self._wait_closed.set_result(True)
        self._on_connection_lost(exc)

    def _on_connection_lost(self, exc):
        pass

    def protocol_error(self, msg):
        self.log.error(msg)
        self.transport.close()

    def sendString(self, data):
        l = len(data)
        if l > self.max_length_send:
            raise ValueError("Data too big")
        header = struct.pack(self.prefix_format, len(data))
        self.transport.write(header)
        self.transport.write(data)

    def ping(self, data):
        raise NotImplementedError()

    def pong(self, data):
        raise NotImplementedError()

    def data_received(self, data):
        self._buffer += data
        pos = 0
        remaining = len(self._buffer)
        while remaining >= self.prefix_length:
            # do not recalculate header if available from previous call
            if self._header:
                frame_type, frame_length = self._header
            else:
                header = self._buffer[pos : pos + self.prefix_length]
                frame_type = ord(header[0:1]) & 0b00000111
                if frame_type > FRAME_TYPE_PONG:
                    self.protocol_error("Invalid frame type")
                    return
                frame_length = struct.unpack(self.prefix_format, b"\0" + header[1:])[0]
                if frame_length > self.max_length:
                    self.protocol_error("Frame too big")
                    return

            if remaining - self.prefix_length >= frame_length:
                self._header = None
                pos += self.prefix_length
                remaining -= self.prefix_length
                data = self._buffer[pos : pos + frame_length]
                pos += frame_length
                remaining -= frame_length

                if frame_type == FRAME_TYPE_DATA:
                    self.stringReceived(data)
                elif frame_type == FRAME_TYPE_PING:
                    self.ping(data)
                elif frame_type == FRAME_TYPE_PONG:
                    self.pong(data)
            else:
                # save heaader
                self._header = frame_type, frame_length
                break

        self._buffer = self._buffer[pos:]

    def stringReceived(self, data):
        raise NotImplementedError()


class RawSocketProtocol(PrefixProtocol):
    def __init__(self):
        # Default receive cap: 16 MB (length exponent 15). The factory overrides
        # this from setProtocolOptions(maxMessagePayloadSize=...) via
        # _set_max_message_size() when the protocol is built.
        self._length_exp = 15
        self.max_length = 2**24

    def _set_max_message_size(self, max_size):
        # Round the configured max up to the next power of two and derive the
        # advertised handshake length exponent (the peer is asked to send
        # messages of at most 2 ** (9 + exp) octets), mirroring the Twisted
        # backend so both enforce and advertise the same receive cap.
        exp = int(math.ceil(math.log(max_size, 2))) - 9
        if exp < 0 or exp > 15:
            raise ValueError("maxMessagePayloadSize must be in [512, 2 ** 24]")
        self._length_exp = exp
        self.max_length = 2 ** (exp + 9)

    def connection_made(self, transport):
        PrefixProtocol.connection_made(self, transport)
        self._handshake_done = False

    def _on_handshake_complete(self):
        raise NotImplementedError()

    def parse_handshake(self):
        buf = bytearray(self._buffer[:4])
        if buf[0] != MAGIC_BYTE:
            raise HandshakeError("Invalid magic byte in handshake")
        ser = buf[1] & 0x0F
        lexp = buf[1] >> 4
        self.max_length_send = 2 ** (lexp + 9)
        if buf[2] != 0 or buf[3] != 0:
            raise HandshakeError("Reserved bytes must be zero")
        return ser, lexp

    def process_handshake(self):
        raise NotImplementedError()

    def data_received(self, data):
        self.log.debug(
            "RawSocker Asyncio: data received {data}", data=_LazyHexFormatter(data)
        )
        if self._handshake_done:
            return PrefixProtocol.data_received(self, data)
        else:
            self._buffer += data
            if len(self._buffer) >= 4:
                try:
                    self.process_handshake()
                except HandshakeError as e:
                    self.protocol_error(f"Handshake error : {e}")
                    return
                self._handshake_done = True
                self._on_handshake_complete()
                data = self._buffer[4:]
                self._buffer = b""
                if data:
                    PrefixProtocol.data_received(self, data)


ERR_SERIALIZER_UNSUPPORTED = 1

ERRMAP = {
    0: "illegal (must not be used)",
    1: "serializer unsupported",
    2: "maximum message length unacceptable",
    3: "use of reserved bits (unsupported feature)",
    4: "maximum connection count reached",
}


class HandshakeError(Exception):
    def __init__(self, msg, code=0):
        Exception.__init__(self, msg if not code else msg + f" : {ERRMAP.get(code)}")


class RawSocketClientProtocol(RawSocketProtocol):
    is_server = False

    def check_serializer(self, ser_id):
        return True

    def process_handshake(self):
        ser_id, err = self.parse_handshake()
        if ser_id == 0:
            raise HandshakeError("Server returned handshake error", err)
        if self.serializer_id != ser_id:
            raise HandshakeError(
                f"Server returned different serializer {ser_id} then requested {self.serializer_id}"
            )

    @property
    def serializer_id(self):
        raise NotImplementedError()

    def connection_made(self, transport):
        RawSocketProtocol.connection_made(self, transport)
        # start handshake
        hs = bytes(
            bytearray([MAGIC_BYTE, self._length_exp << 4 | self.serializer_id, 0, 0])
        )
        transport.write(hs)
        self.log.debug("RawSocket Asyncio: Client handshake sent")


class RawSocketServerProtocol(RawSocketProtocol):
    is_server = True

    def supports_serializer(self, ser_id):
        raise NotImplementedError()

    def process_handshake(self):
        def send_response(lexp, ser_id):
            b2 = lexp << 4 | (ser_id & 0x0F)
            self.transport.write(bytes(bytearray([MAGIC_BYTE, b2, 0, 0])))

        ser_id, _lexp = self.parse_handshake()
        if not self.supports_serializer(ser_id):
            send_response(ERR_SERIALIZER_UNSUPPORTED, 0)
            raise HandshakeError(
                f"Serializer unsupported : {ser_id}"
            )
        send_response(self._length_exp, ser_id)


# this is transport independent part of WAMP protocol
class WampRawSocketMixinGeneral:
    def _on_handshake_complete(self):
        self.log.debug("WampRawSocketProtocol: Handshake complete")
        # RawSocket connection established. Now let the user WAMP session factory
        # create a new WAMP session and fire off session open callback.
        try:
            if self._transport_details.is_secure:
                # now that the TLS opening handshake is complete, the actual TLS channel ID
                # will be available. make sure to set it!
                channel_id = {
                    "tls-unique": transport_channel_id(
                        self.transport, self._transport_details.is_server, "tls-unique"
                    ),
                }
                self._transport_details.channel_id = channel_id

            self._session = self.factory._factory()
            self._session.onOpen(self)
        except Exception as e:
            # Exceptions raised in onOpen are fatal ..
            self.log.warn(
                "WampRawSocketProtocol: ApplicationSession constructor / onOpen raised ({err})",
                err=e,
            )
            self.abort()
        else:
            self.log.info("ApplicationSession started.")

    def stringReceived(self, payload):
        self.log.debug(
            "WampRawSocketProtocol: RX octets: {octets}",
            octets=_LazyHexFormatter(payload),
        )
        try:
            for msg in self._serializer.unserialize(payload):
                self.log.debug("WampRawSocketProtocol: RX WAMP message: {msg}", msg=msg)
                self._session.onMessage(msg)

        except ProtocolError as e:
            self.log.warn(
                "WampRawSocketProtocol: WAMP Protocol Error ({err}) - aborting connection",
                err=e,
            )
            self.abort()

        except Exception as e:
            self.log.warn(
                "WampRawSocketProtocol: WAMP Internal Error ({err}) - aborting connection",
                err=e,
            )
            self.abort()

    def send(self, msg):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.send`
        """
        if self.isOpen():
            self.log.debug(
                "{func}: TX WAMP message: {msg}", func=hltype(self.send), msg=msg
            )
            try:
                payload, _ = self._serializer.serialize(msg)
            except Exception as e:
                # all exceptions raised from above should be serialization errors ..
                raise SerializationError(
                    f"WampRawSocketProtocol: unable to serialize WAMP application payload ({e})"
                )
            else:
                self.sendString(payload)
                self.log.debug(
                    "WampRawSocketProtocol: TX octets: {octets}",
                    octets=_LazyHexFormatter(payload),
                )
        else:
            raise TransportLost()

    def isOpen(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.isOpen`
        """
        return hasattr(self, "_session") and self._session is not None


# this is asyncio dependent part of WAMP protocol
class WampRawSocketMixinAsyncio:
    """
    Base class for asyncio-based WAMP-over-RawSocket protocols.
    """

    def _on_connection_lost(self, exc):
        try:
            wasClean = exc is None
            self._session.onClose(wasClean)
        except Exception as e:
            # silently ignore exceptions raised here ..
            self.log.warn(
                "WampRawSocketProtocol: ApplicationSession.onClose raised ({err})",
                err=e,
            )
        self._session = None

    def close(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.close`
        """
        if self.isOpen():
            self.transport.close()
        else:
            raise TransportLost()

    def abort(self):
        """
        Implements :func:`autobahn.wamp.interfaces.ITransport.abort`
        """
        if self.isOpen():
            if hasattr(self.transport, "abort"):
                # ProcessProtocol lacks abortConnection()
                self.transport.abort()
            else:
                self.transport.close()
        else:
            raise TransportLost()


@public
class WampRawSocketServerProtocol(
    WampRawSocketMixinGeneral, WampRawSocketMixinAsyncio, RawSocketServerProtocol
):
    """
    asyncio-based WAMP-over-RawSocket server protocol.

    Implements:

        * :class:`autobahn.wamp.interfaces.ITransport`
    """

    def supports_serializer(self, ser_id):
        if ser_id in self.factory._serializers:
            self._serializer = copy.copy(self.factory._serializers[ser_id])
            self.log.debug(
                "WampRawSocketProtocol: client wants to use serializer '{serializer}'",
                serializer=ser_id,
            )
            return True
        else:
            self.log.debug(
                "WampRawSocketProtocol: opening handshake - no suitable serializer found (client requested {serializer}, and we have {serializers}",
                serializer=ser_id,
                serializers=self.factory._serializers.keys(),
            )
            self.abort()
            return False


@public
class WampRawSocketClientProtocol(
    WampRawSocketMixinGeneral, WampRawSocketMixinAsyncio, RawSocketClientProtocol
):
    """
    asyncio-based WAMP-over-RawSocket client protocol.

    Implements:

        * :class:`autobahn.wamp.interfaces.ITransport`
    """

    @property
    def serializer_id(self):
        if not hasattr(self, "_serializer"):
            self._serializer = copy.copy(self.factory._serializer)
        return self._serializer.RAWSOCKET_SERIALIZER_ID


class WampRawSocketFactory:
    """
    Adapter class for asyncio-based WebSocket client and server factories.def dataReceived(self, data):
    """

    log = txaio.make_logger()

    # RawSocket max payload size is 16M
    # (https://wamp-proto.org/_static/gen/wamp_latest_ietf.html#handshake)
    _max_message_size = 2**24

    def resetProtocolOptions(self):
        self._max_message_size = 2**24

    def setProtocolOptions(self, maxMessagePayloadSize=None):
        """
        Set RawSocket protocol options. Mirrors the Twisted RawSocket factory so
        the same ``maxMessagePayloadSize`` knob configures the receive size limit
        on both backends.

        :param maxMessagePayloadSize: Maximum length (in octets) of a received
            RawSocket message, in ``[512, 2**24]``; rounded up to the next power
            of two for the advertised handshake length exponent. ``None`` leaves
            the current value unchanged (default ``2**24`` = 16 MB).
        """
        self.log.debug(
            "{klass}.setProtocolOptions(maxMessagePayloadSize={maxMessagePayloadSize})",
            klass=self.__class__.__name__,
            maxMessagePayloadSize=maxMessagePayloadSize,
        )
        assert maxMessagePayloadSize is None or (
            isinstance(maxMessagePayloadSize, int)
            and maxMessagePayloadSize >= 512
            and maxMessagePayloadSize <= 2**24
        )
        if (
            maxMessagePayloadSize is not None
            and maxMessagePayloadSize != self._max_message_size
        ):
            self._max_message_size = maxMessagePayloadSize

    @public
    def __call__(self):
        proto = self.protocol()
        proto.factory = self
        proto._set_max_message_size(self._max_message_size)
        return proto


@public
class WampRawSocketServerFactory(WampRawSocketFactory):
    """
    asyncio-based WAMP-over-RawSocket server protocol factory.
    """

    protocol = WampRawSocketServerProtocol

    def __init__(self, factory, serializers=None):
        """

        :param factory: A callable that produces instances that implement
            :class:`autobahn.wamp.interfaces.ITransportHandler`
        :type factory: callable

        :param serializers: A list of WAMP serializers to use (or ``None``
            for all available serializers).
        :type serializers: list of objects implementing
            :class:`autobahn.wamp.interfaces.ISerializer`
        """
        if callable(factory):
            self._factory = factory
        else:
            self._factory = lambda: factory

        # when no serializers were requested specifically, then support
        # all that are available
        if serializers is None:
            serializers = [serializer() for serializer in get_serializers()]

            if not serializers:
                raise Exception("could not import any WAMP serializers")

        self._serializers = {ser.RAWSOCKET_SERIALIZER_ID: ser for ser in serializers}


@public
class WampRawSocketClientFactory(WampRawSocketFactory):
    """
    asyncio-based WAMP-over-RawSocket client factory.
    """

    protocol = WampRawSocketClientProtocol

    def __init__(self, factory, serializer=None):
        """

        :param factory: A callable that produces instances that implement
            :class:`autobahn.wamp.interfaces.ITransportHandler`
        :type factory: callable

        :param serializer: The WAMP serializer to use (or ``None`` for
           "best" serializer, chosen as the first serializer available from
           this list: CBOR, MessagePack, UBJSON, JSON).
        :type serializer: object implementing :class:`autobahn.wamp.interfaces.ISerializer`
        """
        if callable(factory):
            self._factory = factory
        else:
            self._factory = lambda: factory

        # when no serializer was requested specifically, use the first
        # one available
        if serializer is None:
            serializers = get_serializers()
            if serializers:
                serializer = serializers[0]()

        if serializer is None:
            raise Exception("could not import any WAMP serializer")

        self._serializer = serializer



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os

if os.environ.get("USE_TWISTED", False):
    from autobahn.twisted import wamp
    from autobahn.twisted.util import sleep
    from twisted.application import service
    from twisted.internet import defer
    from twisted.trial import unittest

    class CaseComponent(wamp.ApplicationSession):
        """
        Application code goes here. This is an example component that calls
        a remote procedure on a WAMP peer, subscribes to a topic to receive
        events, and then stops the world after some events.
        """

        def __init__(self, config):
            wamp.ApplicationSession.__init__(self, config)
            self.test = config.extra["test"]
            self.stop = False
            self._logline = 1
            self.finished = False

        def log(self, *args):
            if len(args) > 1:
                sargs = ", ".join(str(s) for s in args)
            elif len(args) == 1:
                sargs = args[0]
            else:
                sargs = "-"

            msg = f"= : {self._logline:>3} : {self.__class__.__name__:<20} : {sargs}"
            self._logline += 1
            print(msg)

        def finish(self):
            if not self.finished:
                self.test.deferred.callback(None)
                self.finished = True
            else:
                print("already finished")

    class Case1_Backend(CaseComponent):
        @defer.inlineCallbacks
        def onJoin(self, details):
            self.log("joined")

            def add2(x, y):
                self.log(f"add2 invoked: {x}, {y}")
                return x + y

            yield self.register(add2, "com.mathservice.add2")
            self.log("add2 registered")

            self.finish()

    class Case1_Frontend(CaseComponent):
        @defer.inlineCallbacks
        def onJoin(self, details):
            self.log("joined")

            try:
                res = yield self.call("com.mathservice.add2", 2, 3)
            except Exception as e:
                self.log(f"call error: {e}")
            else:
                self.log(f"call result: {res}")

            self.finish()

    class Case2_Backend(CaseComponent):
        @defer.inlineCallbacks
        def onJoin(self, details):
            self.log("joined")

            def ping():
                self.log("ping() is invoked")
                return

            def add2(a, b):
                self.log("add2() is invoked", a, b)
                return a + b

            def stars(nick="somebody", stars=0):
                self.log("stars() is invoked", nick, stars)
                return f"{nick} starred {stars}x"

            def orders(product, limit=5):
                self.log("orders() is invoked", product, limit)
                return [f"Product {i}" for i in range(50)][:limit]

            def arglen(*args, **kwargs):
                self.log("arglen() is invoked", args, kwargs)
                return [len(args), len(kwargs)]

            yield self.register(ping, "com.arguments.ping")
            yield self.register(add2, "com.arguments.add2")
            yield self.register(stars, "com.arguments.stars")
            yield self.register(orders, "com.arguments.orders")
            yield self.register(arglen, "com.arguments.arglen")

            self.log("procedures registered")

    class Case2_Frontend(CaseComponent):
        @defer.inlineCallbacks
        def onJoin(self, details):
            self.log("joined")

            yield sleep(1)

            yield self.call("com.arguments.ping")
            self.log("Pinged!")

            res = yield self.call("com.arguments.add2", 2, 3)
            self.log(f"Add2: {res}")

            starred = yield self.call("com.arguments.stars")
            self.log(f"Starred 1: {starred}")

            starred = yield self.call("com.arguments.stars", nick="Homer")
            self.log(f"Starred 2: {starred}")

            starred = yield self.call("com.arguments.stars", stars=5)
            self.log(f"Starred 3: {starred}")

            starred = yield self.call("com.arguments.stars", nick="Homer", stars=5)
            self.log(f"Starred 4: {starred}")

            orders = yield self.call("com.arguments.orders", "coffee")
            self.log(f"Orders 1: {orders}")

            orders = yield self.call("com.arguments.orders", "coffee", limit=10)
            self.log(f"Orders 2: {orders}")

            arglengths = yield self.call("com.arguments.arglen")
            self.log(f"Arglen 1: {arglengths}")

            arglengths = yield self.call("com.arguments.arglen", 1, 2, 3)
            self.log(f"Arglen 1: {arglengths}")

            arglengths = yield self.call("com.arguments.arglen", a=1, b=2, c=3)
            self.log(f"Arglen 2: {arglengths}")

            arglengths = yield self.call("com.arguments.arglen", 1, 2, 3, a=1, b=2, c=3)
            self.log(f"Arglen 3: {arglengths}")

            self.log("finishing")

            self.finish()

    class TestRpc(unittest.TestCase):
        if os.environ.get("WAMP_ROUTER_URL") is None:
            skip = (
                "Please provide WAMP_ROUTER_URL environment with url to "
                "WAMP router to run WAMP integration tests"
            )

        def setUp(self):
            self.url = os.environ.get("WAMP_ROUTER_URL")
            self.realm = "realm1"

        @defer.inlineCallbacks
        def runOneTest(self, components):
            self.deferred = defer.Deferred()
            app = service.MultiService()
            for component in components:
                c = wamp.Service(
                    url=self.url,
                    extra=dict(test=self),
                    realm=self.realm,
                    make=component,
                )
                c.setServiceParent(app)

            app.startService()
            yield self.deferred
            app.stopService()

        @defer.inlineCallbacks
        def test_case1(self):
            yield self.runOneTest([Case1_Backend, Case1_Frontend])

        @defer.inlineCallbacks
        def test_case2(self):
            yield self.runOneTest([Case2_Backend, Case2_Frontend])
